/**
 * @file reed_muller.c
 * Constant time implementation of Reed-Muller code RM(1,7)
 */

// produces an encode routine with fewer instructions, but bigger
#undef BIGENCODE
// produces the same code, but is more generic
#undef GCCVECTORS

#include "reed_muller.h"
#include "parameters.h"
#include <stdint.h>
#include <string.h>
#include <x86intrin.h>

// number of repeated code words
#define MULTIPLICITY                   CEIL_DIVIDE(PARAM_N2, 128)

// codeword is 128 bits, seen multiple ways
typedef union {
	__mmask32 mask[4];
	uint8_t u8[16];
	uint32_t u32[4];
	uint64_t u64[2];
} codeword __attribute__ ((aligned (16)));

// a 512 bit vector of 32 ints of 16 bits
typedef int16_t i16x32 __attribute__ ((vector_size (64)));

// Expanded codeword is 16*128 bits, seen multiple ways
typedef union {
	__m512i mm[4];     // four intel vectors
	i16x32 v[4];       // four vectors of 32 ints
	int16_t i16[128];
} expandedCodeword __attribute__ ((aligned (64))) ;

// copy bit 0 into all bits of a 64 bit value
#define BIT0MASK(x) (int64_t)(-((x) & 1))

// vectors for the hadamard transform
static const i16x32 even_indices = {
	0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26, 28, 30,
	32, 34, 36, 38, 40, 42, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62,
};
static const i16x32 odd_indices = {
	1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29, 31,
	33, 35, 37, 39, 41, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63,
};


void encode(codeword *word, int32_t message);
void expand_and_sum(expandedCodeword *dst, codeword src[]);
void hadamard(expandedCodeword *src, expandedCodeword *dst);
int32_t find_peaks(expandedCodeword *transform);



/**
 * @brief Encode a single byte into a single codeword using RM(1,7)
 *
 * Encoding matrix of this code:
 * bit pattern (note that bits are numbered big endian)
 * 0   aaaaaaaa aaaaaaaa aaaaaaaa aaaaaaaa
 * 1   cccccccc cccccccc cccccccc cccccccc
 * 2   f0f0f0f0 f0f0f0f0 f0f0f0f0 f0f0f0f0
 * 3   ff00ff00 ff00ff00 ff00ff00 ff00ff00
 * 4   ffff0000 ffff0000 ffff0000 ffff0000
 * 5   00000000 ffffffff 00000000 ffffffff
 * 6   00000000 00000000 ffffffff ffffffff
 * 7   ffffffff ffffffff ffffffff ffffffff
 *
 * @param[out] word An RM(1,7) codeword
 * @param[in] message A message to encode
 */
inline void encode(codeword *word, int32_t message) {
	#ifdef BIGENCODE
		// the two parts of the word are identical
		// except for encoding of bit 6
		uint64_t first_word;
		// bit 7 flips all the bits, do that first to save work
		first_word = BIT0MASK(message >> 7);
		// bits 0, 1, 2, 3, 4,5 are the same for both values
		// (Warning: in the bit matrix above, low bits are at the left!)
		first_word ^= BIT0MASK(message >> 0) & 0xaaaaaaaaaaaaaaaa;
		first_word ^= BIT0MASK(message >> 1) & 0xcccccccccccccccc;
		first_word ^= BIT0MASK(message >> 2) & 0xf0f0f0f0f0f0f0f0;
		first_word ^= BIT0MASK(message >> 3) & 0xff00ff00ff00ff00;
		first_word ^= BIT0MASK(message >> 4) & 0xffff0000ffff0000;
		first_word ^= BIT0MASK(message >> 5) & 0xffffffff00000000;
		// we can store this in the first quarter
		word->u64[0] = first_word;
		// bit 6 flips last bits
		first_word ^= BIT0MASK(message >> 6);
		word->u64[1] = first_word;
	#else
		// the four parts of the word are identical
		// except for encoding bits 5 and 6
		int32_t first_word;
		// bit 7 flips all the bits, do that first to save work
		first_word = BIT0MASK(message >> 7);
		// bits 0, 1, 2, 3, 4 are the same for all four longs
		// (Warning: in the bit matrix above, low bits are at the left!)
		first_word ^= BIT0MASK(message >> 0) & 0xaaaaaaaa;
		first_word ^= BIT0MASK(message >> 1) & 0xcccccccc;
		first_word ^= BIT0MASK(message >> 2) & 0xf0f0f0f0;
		first_word ^= BIT0MASK(message >> 3) & 0xff00ff00;
		first_word ^= BIT0MASK(message >> 4) & 0xffff0000;
		// we can store this in the first quarter
		word->u32[0] = first_word;
		// bit 5 flips entries 1 and 3; bit 6 flips 2 and 3
		first_word ^= BIT0MASK(message >> 5);
		word->u32[1] = first_word;
		first_word ^= BIT0MASK(message >> 6);
		word->u32[3] = first_word;
		first_word ^= BIT0MASK(message >> 5);
		word->u32[2] = first_word;
	#endif
	return;
}



/**
 * @brief Add multiple codewords into expanded codeword
 *
 * Note: this does not write the codewords as -1 or +1 as the green machine does
 * instead, just 0 and 1 is used.
 * The resulting hadamard transform has:
 * all values are halved
 * the first entry is 64 too high
 *
 * @param[out] dst Structure that contain the expanded codeword
 * @param[in] src Structure that contain the codeword
 */
inline void expand_and_sum(expandedCodeword *dst, codeword src[]) {
	// start converting the first copy
	for (size_t part = 0 ; part < 4 ; part++) {
		dst->mm[part] = _mm512_maskz_set1_epi16(src[0].mask[part], 1);
	}
	// sum the rest of the copies
	for (size_t copy = 1; copy < MULTIPLICITY; copy++) {
		for (size_t part = 0 ; part < 4 ; part++) {
			dst->v[part] += (i16x32) _mm512_maskz_set1_epi16(src[copy].mask[part], 1);
		}
	}
}



/**
 * @brief Hadamard transform
 *
 * Perform hadamard transform of src and store result in dst
 * src is overwritten: it is also used as intermediate buffer
 * Method is best explained if we use H(3) instead of H(7):
 *
 * The routine multiplies by the matrix H(3):
 *                     [1  1  1  1  1  1  1  1]
 *                     [1 -1  1 -1  1 -1  1 -1]
 *                     [1  1 -1 -1  1  1 -1 -1]
 * [a b c d e f g h] * [1 -1 -1  1  1 -1 -1  1] = result of routine
 *                     [1  1  1  1 -1 -1 -1 -1]
 *                     [1 -1  1 -1 -1  1 -1  1]
 *                     [1  1 -1 -1 -1 -1  1  1]
 *                     [1 -1 -1  1 -1  1  1 -1]
 * You can do this in three passes, where each pass does this:
 * set lower half of buffer to pairwise sums,
 * and upper half to differences
 * index     0        1        2        3        4        5        6        7
 * input:    a,       b,       c,       d,       e,       f,       g,       h
 * pass 1:   a+b,     c+d,     e+f,     g+h,     a-b,     c-d,     e-f,     g-h
 * pass 2:   a+b+c+d, e+f+g+h, a-b+c-d, e-f+g-h, a+b-c-d, e+f-g-h, a-b-c+d, e-f-g+h
 * pass 3:   a+b+c+d+e+f+g+h   a+b-c-d+e+f-g-h   a+b+c+d-e-f-g-h   a+b-c-d-e+-f+g+h
 * a-b+c-d+e-f+g-h   a-b-c+d+e-f-g+h   a-b+c-d-e+f-g+h   a-b-c+d-e+f+g-h
 * This order of computation is chosen because it vectorises well.
 * Likewise, this routine multiplies by H(7) in seven passes.
 *
 * @param[out] src Structure that contain the expanded codeword
 * @param[out] dst Structure that contain the expanded codeword
 */
inline void hadamard(expandedCodeword *src, expandedCodeword *dst) {
	// the passes move data:
	// src -> dst -> src -> dst -> src -> dst -> src -> dst
	// using p1 and p2 alternately
	expandedCodeword *p1 = src;
	expandedCodeword *p2 = dst;
	for (size_t pass = 0 ; pass < 7 ; pass++) {
		#ifdef GCCVECTORS
			i16x32 odd, even;
			even = __builtin_shuffle(p1->v[0], p1->v[1], even_indices);
			odd = __builtin_shuffle(p1->v[0], p1->v[1], odd_indices);
			p2->v[0] = even + odd;
			p2->v[2] = even - odd;
			even = __builtin_shuffle(p1->v[2], p1->v[3], even_indices);
			odd = __builtin_shuffle(p1->v[2], p1->v[3], odd_indices);
			p2->v[1] = even + odd;
			p2->v[3] = even - odd;
		#else
			__m512i even, odd;
			even = _mm512_permutex2var_epi16(p1->mm[0], (__m512i)even_indices, p1->mm[1]);
			odd = _mm512_permutex2var_epi16(p1->mm[0], (__m512i)odd_indices, p1->mm[1]);
			p2->mm[0] = _mm512_add_epi16(even, odd);
			p2->mm[2] = _mm512_sub_epi16(even, odd);
			even = _mm512_permutex2var_epi16(p1->mm[2], (__m512i)even_indices, p1->mm[3]);
			odd = _mm512_permutex2var_epi16(p1->mm[2], (__m512i)odd_indices, p1->mm[3]);
			p2->mm[1] = _mm512_add_epi16(even, odd);
			p2->mm[3] = _mm512_sub_epi16(even, odd);
		#endif
		// swap p1, p2 for next round
		expandedCodeword *p3 = p1;
		p1 = p2;
		p2 = p3;
	}
}



/**
 * @brief Finding the location of the highest value
 *
 * This is the final step of the green machine: find the location of the highest value,
 * and add 128 if the peak is positive
 * Notes on decoding
 * The standard "Green machine" decoder words as follows:
 * if the received codeword is W, compute (2 * W - 1) * H7
 * The entries of the resulting vector are always even and vary from
 * -128 (= the complement is a code word, add bit 7 to decode)
 * via 0 (this is a different codeword)
 * to 128 (this is the code word).
 *
 * Our decoding differs in two ways:
 * - We take W instead of 2 * W - 1 (so the entries are 0,1 instead of -1,1)
 * - We take the sum of the repititions (so the entries are 0..MULTIPLICITY)
 * This implies that we have to subtract 64M (M=MULTIPLICITY)
 * from the first entry to make sure the first codewords is handled properly
 * and that the entries vary from -64M to 64M.
 * -64M or 64M stands for a perfect codeword.
 * If there are fewer than 32M errors, there is always a unique codeword
 * which an entry with absolute value > 32M;
 * this is because an error changes an entry by 1.
 * The highest number that seem to be decodable is 50 errors, so that the
 * highest entries in the hadamard transform can be as low as 12.
 * But this is different for the repeated code.
 * Because multiple codewords are added, this changes: the lowest value of the
 * hadamard transform of the sum of six words is seen to be as low as 43 (!),
 * which is way less than 12*6.
 *
 * It is possible that there are more errors, but the word is still uniquely
 * decodable: we found a word with distance of 50 from the nearest codeword.
 * That means that the highest entry can be as low as 14M.
 * Since we have to do binary search, we search for the range 1-64M
 * which can be done in 6+l2g(M) steps.
 * The binary search is based on (values>32M are unique):
 * M  32M     min>  max>  firstStep #steps
 * 2   64       1   64    33 +- 16    6
 * 4  128       1  128    65 +- 32    7
 * 6  192       1  192   129 +- 64    8
 *
 * As a check, we run a sample for M=6 to see the peak value; it ranged
 * from 43 to 147, so my analysis looks right. Also, it shows that decoding
 * far beyond the bound of 32M is needed.
 *
 * For the vectors, it would be tempting to use 8 bit ints,
 * because the values "almost" fit in there.
 * We could use some trickery to fit it in 8 bits, like saturated add or
 * division by 2 in a late step.
 * Unfortunately, these instructions do not exist.
 * the adds _mm512_adds_epi8 is available only on the latest processors,
 * and division, shift, mulhi are not available at all for 8 bits.
 * So, we use 16 bit ints.
 *
 * For the search of the optimal comparison value,
 * remember the transform contains 64M-d,
 * where d are the distances to the codewords.
 * The highest value gives the most likely codeword.
 * There is not fast vectorized way to find this value, so we search for the
 * maximum value itself.
 * In each pass, we collect a bit map of the transform values that are,
 * say >bound.  There are three cases:
 * bit map = 0: all code words are further away than 64M-bound (decrease bound)
 * bit map has one bit: one unique code word has distance < 64M-bound
 * bit map has multiple bits: multiple words (increase bound)
 * We will search for the lowest value of bound that gives a nonzero bit map.
 *
 * @param[in] transform Structure that contain the expanded codeword
 */
inline int32_t find_peaks(expandedCodeword *transform) {
	// a whole lot of vector variables
	__m512i abs_rows[4], bound, active_row, max_abs_rows;
	size_t bitmap;
	// compute absolute value of transform
	for (size_t i = 0 ; i < 4 ; i++) {
		abs_rows[i] = _mm512_abs_epi16(transform->mm[i]);
	}
	// compute a vector of 32 elements which contains the maximum somewhere
	// (later used to compute bits 0 through 4 of message)
	max_abs_rows = abs_rows[0];
	for (size_t i = 1 ; i < 4 ; i++) {
		max_abs_rows = _mm512_max_epi16(max_abs_rows, abs_rows[i]);
	}

	// do binary search for the highest value that is lower than the maximum
	// loop invariant: lower gives bit map = 0, lower + width gives bit map > 0
	int32_t lower = 1;
	// this gives 64, 128 or 256 for MULTIPLICITY = 2, 4, 6
	int32_t width = 1 << (5 + MULTIPLICITY / 2);
	// if you don't unroll this loop, it fits in the loop cache
	// uncomment the line below to speeding up the program by a few percent
	// #pragma GCC unroll 0
	while (width > 1) {
		width >>= 1;
		// compare with lower + width; put result in bitmap
		// make vector from value of new bound
		bound = _mm512_broadcastw_epi16(_mm_cvtsi32_si128(lower + width));
		bitmap = _mm512_cmpgt_epi16_mask(max_abs_rows, bound);
		// step up if there are any matches
		// rely on compiler to use conditional move here
		int32_t step_mask = (bitmap == 0) - 1;
		lower += step_mask & width;
	}
	// lower+width contains the maximum value of the vector
	// or less, if the maximum is very high (which is OK)
	// normally, there is one maximum, but sometimes there are more
	// find where the maxima occur in the maximum vector
	// (each determines lower 5 bits of peak position)
	// construct vector filled with bound
	bound = _mm512_broadcastw_epi16(_mm_cvtsi32_si128(lower + width));

	// find in which of the 8 groups a maximum occurs
	// to compute bits 5, 6 of message
	// find lowest value by searching backwards
	// skip first check to save time
	int32_t message = 0x60;
	for (int32_t i = 3 ; i >= 0 ; i--) {
		bitmap = _mm512_cmpge_epi16_mask(abs_rows[i], bound);
		message = (bitmap == 0) ? message : i << 5;
	}
	// we decided which row of the matrix contains the lowest match
	// select proper row
	active_row = abs_rows[message >> 5];

	// get the column number of the vector element
	size_t column_map = _mm512_cmpge_epi16_mask(active_row, bound);
	message |= _bit_scan_forward(column_map);

	// set bit 7 if sign of biggest value is positive
	// make sure a jump isn't generated by the compiler
	message |= (0x8000 & ~transform->i16[message]) >> 8;
	return message;
}



/**
 * @brief Encodes the received word
 *
 * The message consists of N1 bytes each byte is encoded into PARAM_N2 bits,
 * or MULTIPLICITY repeats of 128 bits
 *
 * @param[out] cdw Array of size VEC_N1N2_SIZE_64 receiving the encoded message
 * @param[in] msg Array of size VEC_N1_SIZE_64 storing the message
 */
void reed_muller_encode(uint64_t *cdw, const uint64_t *msg) {
	uint8_t *message_array = (uint8_t *) msg;
	codeword *codeArray = (codeword *) cdw;
	for (size_t i = 0 ; i < VEC_N1_SIZE_BYTES ; i++) {
		// fill entries i * MULTIPLICITY to (i+1) * MULTIPLICITY
		int32_t pos = i * MULTIPLICITY;
		// encode first word
		encode(&codeArray[pos], message_array[i]);
		// copy to other identical codewords
		for (size_t copy = 1 ; copy < MULTIPLICITY ; copy++) {
			memcpy(&codeArray[pos + copy], &codeArray[pos], sizeof(codeword));
		}
	}
	return;
}



/**
 * @brief Decodes the received word
 *
 * Decoding uses fast hadamard transform, for a more complete picture on Reed-Muller decoding, see MacWilliams, Florence Jessie, and Neil James Alexander Sloane.
 * The theory of error-correcting codes codes @cite macwilliams1977theory
 *
 * @param[out] msg Array of size VEC_N1_SIZE_64 receiving the decoded message
 * @param[in] cdw Array of size VEC_N1N2_SIZE_64 storing the received word
 */
void reed_muller_decode(uint64_t *msg, const uint64_t *cdw) {
	uint8_t *message_array = (uint8_t *) msg;
	codeword *codeArray = (codeword *) cdw;
	expandedCodeword expanded;
	for (size_t i = 0 ; i < VEC_N1_SIZE_BYTES ; i++) {
		// collect the codewords
		expand_and_sum(&expanded, &codeArray[i * MULTIPLICITY]);
		// apply hadamard transform
		expandedCodeword transform;
		hadamard(&expanded, &transform);
		// fix the first entry to get the half Hadamard transform
		transform.i16[0] -= 64 * MULTIPLICITY;
		// finish the decoding
		message_array[i] = find_peaks(&transform);
	}
}
