/**
 * \file gf2x.c
 * \brief AVX2 implementation of multiplication of two polynomials
 */

#include "gf2x.h"
#include "rng.h"
#include "parameters.h"
#include <stdint.h>
#include <string.h>
#include <immintrin.h>

#define VEC_N_ARRAY_SIZE_VEC CEIL_DIVIDE(PARAM_N, 256) /*!< The number of needed vectors to store PARAM_N bits*/
#define TABLE 16
#define WORD 64
#define LAST64 (PARAM_N >> 6)

uint64_t a1_times_a2[VEC_N_ARRAY_SIZE_VEC << 3];
uint64_t tmp_reduce[VEC_N_ARRAY_SIZE_VEC << 2];
__m256i *o256 = (__m256i *) tmp_reduce;

static inline void swap(uint16_t *tab, uint16_t elt1, uint16_t elt2);
static inline void reduce(uint64_t *o, uint64_t *a);
static inline void fast_convolution_mult(uint64_t *o, const  uint32_t *a1, const uint64_t *a2, const uint16_t weight, AES_XOF_struct *ctx);


/**
 * @brief swap two elements in a table
 *
 * This function exchanges tab[elt1] with tab[elt2]
 *
 * @param[in] tab Pointer to the table
 * @param[in] elt1 , index of the first element
 * @param[in] elt2, index of the second element
 */
static inline void swap(uint16_t *tab, uint16_t elt1, uint16_t elt2) {
	int32_t sw = tab[elt1];

	tab[elt1] = tab[elt2];
	tab[elt2] = sw;
}



/**
 * @brief Compute o(x) = a(x) mod \f$ X^n - 1\f$
 *
 * This function computes the modular reduction of the polynomial a(x)
 *
 * @param[out] o Pointer to the result
 * @param[in] a Pointer to the polynomial a(x)
 */
static inline void reduce(uint64_t *o, uint64_t *a) {
	__m256i r256, carry256;
	__m256i *a256 = (__m256i *) a;
	static const int32_t dec64 = PARAM_N & 0x3f;
	static const int32_t d0 = WORD - dec64;
	int32_t i, i2;

	for (i = LAST64 ; i < (PARAM_N >> 5) - 4 ; i += 4) {
		r256 = _mm256_lddqu_si256((__m256i const *) (& a[i]));
		r256 = _mm256_srli_epi64(r256, dec64);
		carry256 = _mm256_lddqu_si256((__m256i const *) (& a[i + 1]));
		carry256 = _mm256_slli_epi64(carry256, d0);
		r256 ^= carry256;
		i2 = (i - LAST64) >> 2;
		o256[i2] = a256[i2] ^ r256;
	}

	r256 = _mm256_lddqu_si256((__m256i const *) (& a[i]));
	carry256 = _mm256_lddqu_si256((__m256i const *) (& a[i + 1]));
	r256 = _mm256_srli_epi64(r256, dec64);
	carry256 = _mm256_slli_epi64(carry256, d0);
	r256 ^= carry256;
	i2 = (i - LAST64) >> 2;
	o256[i2] = (a256[i2] ^ r256);
	tmp_reduce[LAST64] &= RED_MASK;
	memcpy(o, tmp_reduce, VEC_N_SIZE_BYTES);
}



/**
 * @brief Compute o(x) = a1(x)a2(x)
 *
 * This function computes product of the polynomial a2(x) with the sparse polynomial a1(x)
 *
 * @param[out] o Pointer to the result
 * @param[in] a1 Pointer to the sparse polynomial a1 (list of degrees of the monomials which appear in a1)
 * @param[in] a2 Pointer to the polynomial a2(x)
 * @param[in] weight Hamming weigth of the sparse polynomial a1(x)
 * @param[in] ctx Pointer to a seed expander used to randomized the multiplication process
 */
static inline void fast_convolution_mult(uint64_t *o, const  uint32_t *a1, const uint64_t *a2, const uint16_t weight, AES_XOF_struct *ctx) {
	int32_t d0, dec16, i16, i256;
	__m256i tmp256, carry256, tabprime;
	__m256i *o256 = (__m256i *) o;
	__m256i zero256 = (__m256i){0x0UL, 0x0UL, 0x0UL, 0x0UL};
	__m256i table[16][VEC_N_ARRAY_SIZE_VEC + 2];
	__m256i tab[2];
	uint16_t permuted_table[TABLE];
	uint16_t permutation_table[TABLE];
	uint16_t permuted_sparse_vect[PARAM_OMEGA_R];
	uint16_t permutation_sparse_vect[PARAM_OMEGA_R];

	for (int32_t j = 0 ; j < VEC_N_ARRAY_SIZE_VEC << 1 ; j++) {
		o256[j] = zero256;
	}

	tab[0] = (__m256i){a2[0], a2[1], a2[2], a2[3]};
	tab[1] = (__m256i){0x0UL, a2[0], a2[1], a2[2]};
	tabprime = (__m256i){a2[(VEC_N_ARRAY_SIZE_VEC << 2) - 1], 0x0UL, 0x0UL, 0x0UL};

	for (int32_t i = 0 ; i < TABLE ; i++) {
		permuted_table[i] = i;
	}

	seedexpander(ctx, (uint8_t *) permutation_table, TABLE << 1);

	for (int32_t i = 0 ; i < TABLE ; i++) {
		swap(permuted_table + i, 0, permutation_table[i] % (TABLE - i));
	}

	for (int32_t i = 0 ; i < 16 ; i++) {
		int32_t j = permuted_table[i];
		table[j][0] = zero256;
		d0= 64 - i;
		tmp256 = tab[0];
		carry256 = tab[1];
		tmp256 = _mm256_slli_epi64(tmp256, i);
		carry256 = _mm256_srli_epi64(carry256, d0);
		tmp256 =_mm256_xor_si256(tmp256, carry256);
		table[j][1] = tmp256;
		__m256i *p = (__m256i *)(a2);
		__m256i *p1 = (__m256i *)(a2 - 1);

		int32_t k;
		for (k = 1 ; k < VEC_N_ARRAY_SIZE_VEC ; ) {
			tmp256 = _mm256_lddqu_si256((__m256i const *) (p + k));
			tmp256 = _mm256_slli_epi64(tmp256, i);
			carry256 = _mm256_lddqu_si256((__m256i const *)(p1 + k));
			carry256 = _mm256_srli_epi64(carry256, d0);
			tmp256 =_mm256_xor_si256(tmp256, carry256);
			table[j][++k] = tmp256;
		}

		carry256 = tabprime;
		carry256 = _mm256_srli_epi64(carry256, d0);
		tmp256 = carry256;
		table[j][++k] = tmp256;
	}

	for (int32_t i = 0 ; i < weight ; i++) {
		permuted_sparse_vect[i] = i;
	}

	seedexpander(ctx, (uint8_t *) permutation_sparse_vect, weight << 1);

	for (int32_t i = 0 ; i < weight ; i++) {
		swap(permuted_sparse_vect + i, 0, permutation_sparse_vect[i] % (weight - i));
	}

	for (int32_t i = 0 ; i < weight ; i++) {
		i16 = (a1[permuted_sparse_vect[i]] & 0xff) >> 4;
		dec16 = a1[permuted_sparse_vect[i]] & 0xf;
		i256 = (a1[permuted_sparse_vect[i]] >> 8);
		uint16_t *table16 = (uint16_t *) (table[permuted_table[dec16]] + 1);
		__m256i *p = (__m256i *) (table16 - i16);
		__m256i *p1 = o256 + i256;

		for (int32_t j = 0 ; j < VEC_N_ARRAY_SIZE_VEC + 1 ; j++) {
			__m256i t256;
			t256 = _mm256_lddqu_si256((__m256i const *) (p + j));
			p1[j] = _mm256_xor_si256(t256, p1[j]);
		}
	}
}



/**
 * @brief Multiply two polynomials modulo \f$ X^n - 1\f$.
 *
 * This functions multiplies a sparse polynomial <b>a1</b> (of Hamming weight equal to <b>weight</b>)
 * and a dense polynomial <b>a2</b>. The multiplication is done modulo \f$ X^n - 1\f$.
 *
 * @param[out] o Pointer to the result
 * @param[in] a1 Pointer to the sparse polynomial
 * @param[in] a2 Pointer to the dense polynomial
 * @param[in] weight Integer that is the weigt of the sparse polynomial
 * @param[in] ctx Pointer to the randomness context
 */
void vect_mul(uint64_t *o, const uint32_t *a1, const uint64_t *a2, const uint16_t weight, AES_XOF_struct *ctx) {
	fast_convolution_mult(a1_times_a2, a1, a2, weight, ctx);
	reduce(o, a1_times_a2);
}
